#include <unistd.h>
#include "iodefine.h"
#include "ivects.h"
#include "tty.h"

/* This isn't as efficient as an interrupt-driven source; there's an
   extra "stop" bit in the transmission stream because we can't set
   the TX data register soon enough.  However, we can't set the baud
   rate for 115200 accurately enough, and the extra stop bit keeps the
   host from getting out of sync.  */

#define CONSOLE SCI1
#define BGA176	0

#define RXBUFSZ 128
static char rxbuf[RXBUFSZ];
static volatile int rxput, rxget, rxcount;

#define TXBUFSZ 256
static char txbuf[TXBUFSZ];
static volatile int txput, txget, txcount;

static long hz = 12000000;

static void __attribute__((interrupt))
sci1_rx_handler (void)
{
  int b;
  b = CONSOLE.RDR;
  rxbuf[rxput++] = b;
  rxput %= RXBUFSZ;
  rxcount ++;
}

static void __attribute__((interrupt))
sci1_tx_handler (void)
{
  int b;
  if (txcount > 0)
    {
      b = txbuf[txget++];
      CONSOLE.TDR = b;
      txget %= TXBUFSZ;
      txcount --;
    }
}

static void
tty_init_port (int baud)
{
  int i;

    /* Disable, internal clock, div-1 */
  CONSOLE.SCR.BYTE = 0x00;

  /* async, 8N1, PCLK (50MHz) */
  CONSOLE.SMR.BYTE = 0x00;

  /* Disable smart card mode */
  CONSOLE.SCMR.BYTE = 0x00;

  /* div-8 mode in BRR */
  CONSOLE.SEMR.BYTE = 0x10;

  /* Calculate baud rate. */
  CONSOLE.BRR = (hz*4) / (16 * baud) - 1;

  for (i=2200; i; --i)
    asm("");

  /* Enable transmitter, receiver, tx/rx interrupts */
  CONSOLE.SCR.BYTE = 0xf0;
}

void
tty_set_hz (long new_hz)
{
  hz = new_hz;
}

void
tty_init (void)
{
  MSTP(SCI1) = 0;

  /* Enable SCI4 input buffer - P3.0 */
  PORT3.ICR.BIT.B0 = 1;
  PORTF.ICR.BIT.B2 = 1;

  /* CTSn are outputs */
  PORT2.DDR.BIT.B7 = 1;
  PORT3.DDR.BIT.B2 = 1;
  PORTF.DDR.BIT.B0 = 1;

#if BGA176
  IOPORT.PFFSCI.BIT.SCI1S = 1;
#endif

  rxput = rxget = rxcount = 0;

  tty_init_port (115200);

  /* SCI1 interrupts */
  _vectors[VECT(SCI1,RXI1)] = sci1_rx_handler;
  IEN(SCI1,RXI1) = 1;
  IPR(SCI1,RXI1) = 6;
  _vectors[VECT(SCI1,TXI1)] = sci1_tx_handler;
  IEN(SCI1,TXI1) = 1;
  IPR(SCI1,TXI1) = 6;
}

void
tty_putc (char c)
{
  int i;
  int old_flags;
  if (c == '\n')
    tty_putc ('\r');
#if 0
  /* TEND = 0:transmitting, 1:done */
  while (! (CONSOLE.SSR.BIT.TEND))
    ;
  CONSOLE.TDR = c;
#else

  old_flags = __builtin_rx_mvfc (0);

  if (txcount == TXBUFSZ)
    {
      PORT7.DR.BIT.B4 = 1;
      __builtin_rx_setpsw ('I');
      while (txcount == TXBUFSZ)
	asm("");
      PORT7.DR.BIT.B4 = 0;
    }

  __builtin_rx_clrpsw ('I');

  if (txcount == 0 && CONSOLE.SSR.BIT.TDRE)
    {
      CONSOLE.TDR = c;
    }
  else
    {
      txbuf[txput++] = c;
      txput %= TXBUFSZ;
      txcount ++;
    }
  __builtin_rx_mvtc (0, old_flags);
#endif
}

void
tty_puts (char *s)
{
  while (*s)
    tty_putc (*s++);
}

int
write (int fd, const void *vbuf, size_t len)
{
  const char *buf = (const char *)vbuf;
  int ilen = len;
  while (len--)
    tty_putc (*buf++);
  /*tty_putc_port (*buf++, fd == 1 ? 1 : 4); */
  return ilen;
}

char
tty_getc ()
{
  char rv;
  while (rxcount == 0)
    ;
  rv = rxbuf[rxget++];
  rxget %= RXBUFSZ;
  rxcount --;
  return rv;
}

int
tty_ready ()
{
  return rxcount;
}

int
tty_idle ()
{
  return !txcount;
}
